/**
 * absolute bare bones minimum example
 * with blinking LED and UART
 */

#include <stdbool.h>
#include <stm32_misc.h>
#include <stm32f401xe.h>


#define LED_PORT        GPIOA
#define LED_PIN         5
#define LED_ON()        GPIO_SET(LED_PORT, LED_PIN);
#define LED_OFF()       GPIO_CLR(LED_PORT, LED_PIN);
#define LED_TGL()       GPIO_TGL(LED_PORT, LED_PIN);

volatile unsigned sys_ticks = 0;

void init_gpio() {
    // activate port A and configure green LED for output
    RCC->AHB1ENR |= RCC_AHB1ENR_GPIOAEN;
    LED_PORT->MODER |= SHIFT2(LED_PIN, GPIO_MODER_OUTPUT);
    LED_PORT->OSPEEDR |= SHIFT2(LED_PIN, GPIO_OSPEEDR_HIGH);
}

void init_comms() {
    // activate port A and map PA2 and PA3 pins to USART2 TX/RX
    RCC->AHB1ENR  |= RCC_AHB1ENR_GPIOAEN;
    GPIOA->MODER  |= SHIFT2(2, GPIO_MODER_ALTERNATE);
    GPIOA->AFR[0] |= SHIFT4(2, 7);
    GPIOA->MODER  |= SHIFT2(3, GPIO_MODER_ALTERNATE);
    GPIOA->AFR[0] |= SHIFT4(3, 7);

    // activate USART2 and configure it for 460800 Baud
    RCC->APB1ENR |= RCC_APB1ENR_USART2EN;
    USART2->CR1 = USART_CR1_UE | USART_CR1_TE | USART_CR1_RE;
    USART2->CR2 = USART_CR2_STOP_1;
    // baud rate: see reference manual section 19.3.4
    // and also the baud rate tables in this section
    //
    // USARTDIV = Clock / (8 * (2 - OVER8) * BAUDRATE)
    // 460800 Baud @ 42 MHz
    //     USARTDIV = 5.6875
    //     mantissa = 5
    //     fraction = 11 / 16
    USART2->BRR = (5 << 4) | 11;
}

/**
 * send a character to the host application
 */
void send_char(char c) {
    while (!(USART2->SR & USART_SR_TXE));
    USART2->DR = c;
}

void send_str(char* str) {
    unsigned int i = 0;
    while(1) {
        char c = str[i++];
        if (c == 0) {
            return;
        } else {
            send_char(c);
        }
    }
}

void delay(unsigned delay_ticks) {
    unsigned start_time = sys_ticks;
    while (sys_ticks - start_time <= delay_ticks);
}

int main() {
    SysTick_Config(SystemCoreClock / 1000);
    init_gpio();
    init_comms();
    while (1) {
        LED_ON();
        delay(500);
        LED_OFF();
        delay(500);
    }
}

void SysTick_Handler() {
    ++sys_ticks;
}
